import torch
from typing import Tuple


class Weighting:

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        cls = globals()[dump["type"]]
        return cls.from_dump(dump)


class Linear(torch.nn.Linear, Weighting):
    """A wrapper around torch.nn.Linear"""

    @classmethod
    def from_dump(cls, dump: dict) -> 'Linear':
        """returns a 'Linear' layer from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.in_features = dump["in features"]
        obj.out_features = dump["out features"]
        obj.weight = torch.nn.Parameter(torch.tensor(dump["weight"],
                                                     dtype=torch.float))
        if dump["bias"] is None:
            obj.bias = None
        else:
            obj.bias = torch.nn.Parameter(torch.tensor(dump["bias"],
                                                       dtype=torch.float))
        return obj

    def __init__(self, in_features: int, out_features: int, bias=True):
        """
        Parameters
        ----------
        in_features : int
            The number of input features
        out_features : int
            The number of output features
        bias : bool
            If False, there is not bias term in the linear equation:
            y = weight*x + bias
        """
        super().__init__(in_features, out_features, bias=bias)

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "in features": self.in_features,
                "out features": self.out_features,
                "weight": self.weight.tolist(),
                "bias": self.bias if self.bias is None else self.bias.tolist()}

    def shape_out(self, shape_in):
        assert shape_in == []
        return []

    def shape_in(self, shape_out):
        assert shape_out == []
        return []

    @property
    def in_channels(self):
        return self.in_features

    @property
    def out_channels(self):
        return self.out_features


class _Convolution(Weighting):
    """A template for batch norm layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> '_Convolution':
        """returns a '_Convolution' layer from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls.__new__(cls)
        torch.nn.__init__(obj)
        obj.in_channels = dump["in channels"]
        obj.out_channels = dump["out channels"]
        obj.kernel_size = dump["kernel size"]
        obj.stride = dump["stride"]
        obj.weight = torch.nn.Parameter(
                        torch.tensor(dump["weight"], dtype=torch.float))
        if dump["bias"] is None:
            obj.bias = None
        else:
            obj.bias = torch.nn.Parameter(
                        torch.tensor(dump["bias"], dtype=torch.float))
        return obj

    @property
    def dump(self) -> dict:
        d = {"type": type(self).__name__,
             "in channels": self.in_channels,
             "out channels": self.out_channels,
             "kernel size": list(self.kernel_size),
             "stride": list(self.stride),
             "weight": self.weight.tolist()}
        if self.bias is None:
            d["bias"] = None
        else:
            d["bias"] = self.bias.tolist()
        return d

    def shape_out(self, shape_in: list) -> list:
        return [int((d - k)/s + 1) for d, k, s in
                zip(shape_in, self.kernel_size, self.stride)]

    def shape_in(self, shape_out: list) -> list:
        return [(d-1)*s + k for d, k, s in
                zip(shape_out, self.kernel_size, self.stride)]


class Conv1d(torch.nn.Conv1d, _Convolution):
    """A wrapper around torch.nn.Conv1d"""

    def __init__(self, in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int = 1,
                 bias: bool = True):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        kernel_size : int
            The size of the convolved kernel
        stride : int
            The stride used for the convolution
        bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        """
        torch.nn.Conv1d.__init__(self, in_channels, out_channels,
                                 kernel_size, stride, bias=bias)


class Conv2d(torch.nn.Conv2d, _Convolution):
    """A wrapper around torch.nn.Conv2d"""

    def __init__(self, in_channels: int,
                 out_channels: int,
                 kernel_size: Tuple[int, int],
                 stride: Tuple[int, int] = (1, 1),
                 bias: bool = True):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels
        out_channels : int
            The number of output channels
        kernel_size : tuple of int
            The (height, width) size of the convolved kernel
        stride : tuple of int
            The (height, width) stride used for the convolution
        bias : bool
            Whether to have a bias in the linear equation:
            y = weight*x + bias
        """
        torch.nn.Conv2d.__init__(self, in_channels, out_channels,
                                 kernel_size, stride, bias=bias)



import math
import torch
import matplotlib.pyplot as plt
from typing import Union, Callable, List
from ..model import Model


class NeuralNetwork(Model):
    """
    Parameters
    ----------
    module : torch.nn.Module
        the underlying torch module of the model
    optimizer : torch.optim
        the optimizer used for training
    """
    ModuleType: type = None

    @classmethod
    def from_dump(cls, dump: dict) -> 'NeuralNetwork':
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        obj.residuals = dump["residuals"]
        obj.module = cls.ModuleType.from_dump(dump["module"])
        obj.optimization_method = dump["optimization method"]
        obj.norm_update_factor = dump["norm update factor"]
        obj.GPU = dump["GPU"]
        obj.learning_rate = dump["learning rate"]
        obj.L1 = dump["L1"]
        obj.L2 = dump["L2"]
        obj._set_state({"grad": dump["gradient"]})
        return obj

    def __init__(self, *args,
                 GPU: bool = False,
                 learning_rate: float = 1.0E-3,
                 norm_update_factor: Union[float, None] = 0.1,
                 optimization_method: str = "Adam",
                 L1: Union[float, None] = None,
                 L2: Union[float, None] = None,
                 **kwargs):
        """
        Parameters
        ----------
        *args : tuple
            The args passed to the constructor of 'self.module'
        GPU : bool
            If True, calculation on GPU is enabled
        learning_rate : float
            The learning rate of the model
        norm_update_factor : float or None
            The update factor used for batch normalization
        optimization_method : str
            The name of the optimization method
        L1 : float or None
            L1 regularization added to the loss function
        L2 : float or None
            L2 regularization added to the loss function
        **kwargs : dict
            The kwargs passed to the constructor of 'self.module'
        """
        self.module = self.ModuleType(*args, **kwargs)
        self.GPU = GPU
        self.optimization_method = optimization_method
        self.learning_rate = learning_rate
        self.norm_update_factor = norm_update_factor
        self.L1 = L1
        self.L2 = L2
        self.residuals = {"training loss": [],
                          "validation loss": [],
                          "epochs": [],
                          "best epoch": None}

    def train(self, training_data: Union[tuple, Callable],
              validation_data: Union[tuple, Callable, None] = None,
              n_epochs: int = 1000,
              patience: int = 100,
              verbose: bool = True,
              L_minibatchs: Union[int, None] = None):
        """
        Trains a neural network model.

        Parameters
        ----------
        training_data : tuple or callable
            The data used to fit the model on.
            A tuple of (x, y[, weights]) or a callable that yields them.
            The type of each element depends on the model kind.
        validation_data : tuple or callable or None
            The data used to test for early stoping.
            Similar to training_data or None
        n_epochs : int
            The maximum number of epochs
        patience : int
            The number of epochs before early stopping
        verbose : bool
            If True the loss are displayed at each epoch
        L_minibatchs : int or None
            Maximum size of the minibatch
            Or None to process the full batch in one go
        """
        self.module.train()
        # Converts training/validation data to tensors
        if not callable(training_data):
            training_data = self.module.data_to_tensor(*training_data)
        if not callable(validation_data) and validation_data is not None:
            validation_data = self.module.data_to_tensor(*validation_data)
        # Initializing
        if self.residuals["best epoch"] is None:
            best_loss = float("inf")
            best_epoch = 0
        else:
            best_epoch = self.residuals["best epoch"]
            i = self.residuals["epochs"].index(best_epoch)
            best_loss = self.residuals["validation loss"][i]
            for v in ["validation loss", "training loss", "epochs"]:
                self.residuals[v] = self.residuals[v][:i+1]
        best_state = self._get_state()
        # trains the model, stops if the user press 'ctrl+c'
        self._training_loop(training_data, validation_data, n_epochs,
                            patience, verbose, L_minibatchs, best_epoch,
                            best_loss, best_state)

    def plot_residuals(self, ax=None, log: bool = True):
        """
        Plot the training and validation data residuals

        Parameters
        ----------
        ax : matplotlib.axes.Axes or None
            The axes to plot on
        log : bool
            If true the y axis is in log scale
        """
        if ax is None:
            f, ax = plt.subplots()
        epochs = self.residuals["epochs"]
        ax.scatter(epochs, self.residuals["training loss"],
                   marker=".",
                   label="training loss")
        ax.scatter(epochs, self.residuals["validation loss"],
                   marker=".",
                   label="validation loss")
        if self.residuals["best epoch"] is not None:
            ax.axvline(self.residuals["best epoch"], color="k")
        if log:
            ax.set_yscale("log")
        ax.set_ylabel("loss")
        ax.set_xlabel("epochs")
        ax.legend()
        f.tight_layout()

    def __call__(self, X):
        """
        Returns the model's evaluation on the given input

        Parameters:
        -----------
        X : Any
            The input X of the model.
            it's type depend on the neural network type.
            see 'help(self.module)'

        Returns
        -------
        Any :
            The returned Y value of the model.
            it's type depends on the neural network type.
            see 'help(self.module)'
        """
        self.module.eval()
        x, _, _ = self.module.data_to_tensor(X, None, None)
        y = self.module(x)
        return self.module.tensor_to_y(y)

    @property
    def GPU(self) -> bool:
        """
        Returns whether the model is evaluated/trained on GPU

        Returns
        -------
        bool :
            whether the GPU mode is enabled
        """
        return self._GPU

    @GPU.setter
    def GPU(self, enable: bool):
        """
        Set whether the model should be evaluated/trained on GPU.
        This is only available with a CUDA capable GPU.

        Returns
        -------
        bool :
            whether the GPU mode is enabled
        """
        if enable and not(torch.cuda.is_available()):
            raise RuntimeError("CUDA is not available on this computer, \
                                can't train on GPU")
        self._GPU = enable
        if enable:
            self.module.device = torch.device("cuda:0")
        else:
            self.module.device = torch.device("cpu")
        self.module.to(self.module.device)

    @property
    def learning_rate(self) -> float:
        """
        Returns the learning rate used for training

        Returns
        -------
        float :
            the learning rate
        """
        if hasattr(self, "_learning_rate"):
            return self._learning_rate
        else:
            return 0

    @learning_rate.setter
    def learning_rate(self, lr: float):
        """
        set the learning rate for the training

        Parameters:
        -----------
        lr : float
            new learning rate
        """
        self._learning_rate = lr
        for g in self.optimizer.param_groups:
            g["lr"] = lr

    @property
    def optimization_method(self) -> str:
        """
        returns the name of the optimization method

        Returns
        -------
        str :
            name of the method
        """
        return self._optimization_method

    @optimization_method.setter
    def optimization_method(self, name: str):
        """
        set the optimization method for training the model.
        must be the name of an optimizer class from 'torch.optim'.

        This also resets the optimization parameters (gradient momentum,
        learning rate decay, ...)

        Parameters
        ----------
        name : str
            the name of the optimization method
        """
        if not hasattr(torch.optim, name):
            available = [n for n in dir(torch.optim) if n[0] != "_"]
            raise ValueError(f"Invalid optimizer '{name}', "
                             f"valid options are: {available}")
        cls = getattr(torch.optim, name)
        self.optimizer = cls(self.module.parameters(), self.learning_rate)
        self._optimization_method = name

    @property
    def norm_update_factor(self):
        """
        Returns the update factor used for batch normalization

        Returns
        -------
        float or None:
            the update factor
        """
        return self._f

    @norm_update_factor.setter
    def norm_update_factor(self, f: Union[float, None]):
        """
        Set the update factor 'f' used for batch normalization
        moment = f*batch_moment + (1-f)*moment
        Where 'moment' are the mean and variance

        f must be between 0. and 1.
        or None to use an averaging method instead

        Parameters
        ----------
        f : float or None
            the update factor
        """
        assert (f is None) or (0. <= f <= 1.)
        for m in self.module.modules():
            if type(m).__name__.startswith("BatchNorm"):
                m.momentum = f
        self._f = f

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "GPU": self.GPU,
                "learning rate": self.learning_rate,
                "norm update factor": self.norm_update_factor,
                "optimization method": self.optimization_method,
                "L1": self.L1,
                "L2": self.L2,
                "residuals": self.residuals,
                "module": self.module.dump,
                "gradient": self._get_state()["grad"]}

    def _get_state(self) -> tuple:
        """
        Returns a snapshot of the model's state

        The 'state_dict' are deep copied otherwise the saved tensors are
        modified along with the network's training

        Returns
        -------
        dict :
            the state of the model
        """
        params = self.module.state_dict(keep_vars=True)
        grads = {k: None if t.grad is None else t.grad.tolist()
                 for k, t in params.items()}
        return {"params": {k: t.tolist() for k, t in params.items()},
                "grad": grads,
                "optim": self.optimizer.state_dict()}

    def _set_state(self, state: tuple):
        """
        Loads a snapshot of the model's state, as returned by '_get_state'

        Parameters
        ----------
        state : dict
            The state of the model
        """
        if "params" in state.keys():
            params = {k: torch.tensor(t, device=self.module.device)
                      for k, t in state["params"].items()}
            self.module.load_state_dict(params)
        if "grad" in state.keys():
            params = self.module.state_dict(keep_vars=True)
            for key in params.keys():
                t = state["grad"][key]
                if t is not None:
                    t = torch.tensor(t, device=self.module.device)
                params[key].grad = t
        if "optim" in state.keys():
            self.optimizer.load_state_dict(state["optim"])

    def _training_loop(self, training_data: tuple,
                       validation_data: Union[tuple, None],
                       n_epochs: int,
                       patience: int,
                       verbose: bool,
                       L_minibatchs: Union[int, None],
                       best_epoch: int,
                       best_loss: float,
                       best_state: tuple):
        """
        Trains the model for a fixed number of epoch,
        or until validation loss has'nt decreased for 'patience' epochs,
        or until the user press 'ctrl+c'

        At each epoch:
            The parameters are updated using the gradient (0 initially)
            The gradient is set back to 0 (otherwise it accumulates)
            The gradient of the loss is evaluated on the training data
            if validation data are provided:
                The validation loss is evaluated
                if the validation loss is inferior to the previous best:
                    the best loss is updated
                    the state of the model is saved
                otherwise if we waited more than 'patience' epochs:
                    interupts training
        The best state and epoch are then saved

        Parameters:
        -----------
        training_data : tuple or Callable
            The data provided to 'self._batch'
        validation_data : tuple or Callbale or None
            Same as 'training_data' or None if not using early stoping
        n_epochs : int
            The number of epochs to performe
        patience : int
            The number of epochs waiting for improvement
            before early stoping
        verbose : bool
            If True prints models info while training
        L_minibatchs : int or None
            The maximum number of items in a minibatchs
            or None to not to use minibatchs
        best_epoch : int
            the epoch of the previous best state
        best_loss : float
            the value of the previous best validation loss
        best_state : tuple
            the snapshot of the model as returned by 'self._get_state'
        """
        try:
            for epoch in range(best_epoch+1, best_epoch+n_epochs+1):
                self.optimizer.step()
                self.optimizer.zero_grad()
                training_loss = self._batch(training_data, L_minibatchs,
                                            train=True)
                if validation_data is not None:
                    validation_loss = self._batch(validation_data,
                                                  L_minibatchs,
                                                  train=False)
                    if validation_loss < best_loss:
                        best_epoch = epoch
                        best_loss = validation_loss
                        best_state = self._get_state()
                    elif (epoch - best_epoch) > patience:
                        break
                else:
                    best_epoch = epoch
                    validation_loss = None
                self.residuals["training loss"].append(training_loss)
                self.residuals["validation loss"].append(validation_loss)
                self.residuals["epochs"].append(epoch)
                if verbose:
                    msg = f"Epoch {epoch}: train={training_loss:.3g}"
                    if validation_loss is not None:
                        msg += f" val={validation_loss:.3g}"
                        if best_epoch != epoch:
                            msg += " - no improvement"
                    print(msg, flush=True)
        except KeyboardInterrupt:
            if verbose:
                print("Training interrupted by the user", flush=True)
            # Trims data in case user interupted in the midle of the loop
            keys = ["validation loss", "training loss", "epochs"]
            L = min(len(self.residuals[key]) for key in keys)
            for key in keys:
                self.residuals[key] = self.residuals[key][:L]
            best_epoch = min(self.residuals["epochs"][-1], best_epoch)
        finally:
            # load the best state
            if validation_data is not None:
                self._set_state(best_state)
            # Save the best epoch
            self.residuals["best epoch"] = best_epoch

    def _batch(self, data: Union[tuple, Callable],
               *args, **kwargs) -> float:
        """
        Process the data by batch (or in one go),
        and returns the average loss.

        Parameters
        ----------
        data : tuple or Callable
            The (X, Y, weights) to evaluate the loss on,
            or a function that yields them by batch.
            'X' and 'Y' are tensors, 'weights' is a list of float or None.
        *args : tuple
            The arguments passed to _minibatch
        **kwargs : dict
            The keyword arguments passed to _minibatch

        Returns
        -------
        float :
            The loss function averaged over the batchs
        """
        if not callable(data):
            return self._minibatch(data, *args, **kwargs)
        else:
            losses = []
            for batch_data in data():
                loss = self._minibatch(
                    self.module.data_to_tensor(*batch_data),
                    *args, **kwargs)
                losses.append(loss)
            return sum(losses)/len(losses)

    def _minibatch(self, batch_data: tuple,
                   L_minibatchs: Union[int, None],
                   train: bool = False) -> float:
        """
        Process the input batch data in minibatchs
        of maximum size 'L_minibatch', and returns the average loss.

        Parameters
        ----------
        batch_data : tuple
            The (X, Y, weights) to evaluate the loss on.
            'X' and 'Y' are tensors, 'weights' is a list of float or None.
        L_minibatch : int or None
            The maximum number of observations in a minibatch.
            If None the whole data is processed in one go.
        train : bool
            If True, the gradient is back propagated

        Returns
        -------
        float :
            The loss function averaged over the minibatchs
        """
        if L_minibatchs is None:
            return self._eval_loss(*batch_data, train=train)
        else:
            X, Y, weights = batch_data
            L_minibatchs = min(max(1, L_minibatchs), len(X))
            n = math.ceil(len(X)/L_minibatchs)
            bounds = [int(i*len(X)/n) for i in range(n+1)]
            losses = [self._eval_loss(X[start:end], Y[start:end],
                                      None if weights is None
                                      else weights[start:end], train=train)
                      for (start, end) in zip(bounds[:-1], bounds[1:])]
            return sum(losses)/len(losses)
            # losses = []
            # for (start, end) in zip(bounds[:-1], bounds[1:]):
            #     x = X[start:end]
            #     y = Y[start:end]
            #     w = None if weights is None else weights[start:end]
            #     losses.append(self._eval_loss(x, y, w, train=train))

    def _eval_loss(self, x: torch.Tensor, y: torch.Tensor,
                   w: Union[List[float], None] = None,
                   train: bool = False) -> torch.Tensor:
        """
        Evaluates the loss on the given input data.

        If 'train' is True, builds the computational graph
        (Slower to compute/more memory used)
        and backpropagate the gradient.

        Parameters
        ----------
        x : torch.Tensor
            observations
        y : torch.Tensor
            target
        w : List of float, or None
            weights
        train : bool
            If True, grad is backpropagated

        Returns
        -------
        float :
            scalar tensor of the evaluated loss
        """
        if train:
            loss = self.module.loss(self.module(x), y, w)
            loss = self._regularization(loss)
            loss.backward()
        else:
            with torch.no_grad():
                self.module.eval()
                loss = self.module.loss(self.module(x), y, w)
                loss = self._regularization(loss)
                self.module.train()
        torch.cuda.empty_cache()
        return float(loss)

    def _regularization(self, loss: torch.Tensor) -> torch.Tensor:
        """
        Add L1 and L2 regularization terms to the loss

        Parameters
        ----------
        loss : torch.Tensor
            the scalar tensor representing the loss

        Returns
        -------
        torch.Tensor :
            the regularized loss
        """
        if self.L1 is not None:
            norm = sum([torch.norm(p, 1)
                        for p in self.module.parameters()])
            loss = loss + self.L1 * norm
        if self.L2 is not None:
            norm = sum([torch.norm(p, 2)
                        for p in self.module.parameters()])
            loss = loss + self.L2 * norm
        return loss



from .dense_classifier import DenseClassifier
from .dense_regressor import DenseRegressor
from .image_classifier import ImageClassifier
from .semantic_segmenter import SemanticSegmenter
from . import layers



import torch
from typing import Union, List
from .activated import Activated, Activated0d, Activated1d, Activated2d


class Dense(torch.nn.Module):
    """
    A Dense layer is a succession of Activated layers
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.layers = torch.nn.ModuleList()
        for layer in dump["layers"]:
            obj.layers.append(Activated.from_dump(layer))

    def __init__(self, in_channels: int,
                 layers: Union[dict, List[dict]],
                 activation: str = "relu",
                 stacked: bool = False,
                 dropout: Union[None, float] = None,
                 padded: bool = True):
        """
        Parameters
        ----------
        in_channels : int
            the number of channels of the input
        layers : dict, or list of dict
            the parameters of each 'Activated' layer in the dense layer
        activation : str
            default 'activation' parameter for 'ActivatedNd' layers
        stacked : bool
            default 'stacked' parameter for 'ActivatedNd' layers
        dropout : None or float
            default 'dropout' parameter for 'ActivatedNd' layers
        padded : bool
            Only for Dense1d and Dense2d
            default 'padded' parameter for 'ActivatedNd' layers
        """
        super().__init__()
        if isinstance(layers, dict):
            layers = [layers]
        self.layers = torch.nn.ModuleList()
        for kwargs in layers:
            kwargs.setdefault("stacked", stacked)
            kwargs.setdefault("dropout", dropout)
            if type(self) != Dense0d:
                kwargs.setdefault("padded", padded)
            layer = self.ActivatedNd(in_channels, **kwargs)
            in_channels = layer.out_channels(in_channels)
            self.layers.append(layer)

    def forward(self, X):
        for layer in self.layers:
            X = layer(X)
        return X

    def shape_in(self, shape_out: list) -> list:
        shape = shape_out
        for layer in self.layers[::-1]:
            shape = layer.shape_in(shape)
        return shape

    def shape_out(self, shape_in: list) -> list:
        shape = shape_in
        for layer in self.layers:
            shape = layer.shape_out(shape)
        return shape

    def in_channels(self, out_channels: int) -> int:
        channels = out_channels
        for layer in self.layers[::-1]:
            channels = layer.in_channels(channels)
        return channels

    def out_channels(self, in_channels: int) -> int:
        channels = in_channels
        for layer in self.layers:
            channels = layer.out_channels(channels)
        return channels

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "layers": [layer.dump for layer in self.layers]}


class Dense0d(Dense):

    ActivatedNd = Activated0d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Dense1d(Dense):

    ActivatedNd = Activated1d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Dense2d(Dense):

    ActivatedNd = Activated2d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import torch
from typing import Union, List, Tuple
from .encoder import Encoder1d, Encoder2d
from .decoder import Decoder1d, Decoder2d


class UNet(torch.nn.Module):
    """
    An UNet structure is a encoder followed by a decoder
    with a skip connection between layers of same depth
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        cls = globals()[dump["type"]]
        return cls.from_dump(dump)

    def __init__(self, in_channels: int,
                 downsampling: List[Union[dict, List[dict]]],
                 pooling: List[Union[int, Tuple[int, int]]],
                 upsampling: List[Union[dict, List[dict]]],
                 pooling_type: str = "max",
                 upsampling_method: str = "nearest",
                 activation: str = "relu",
                 stacked: bool = False,
                 dropout: Union[float, None] = None):
        """
        in_channels : int
            The number of channels of the input
        downsampling : list of [dict / list of dict]
            the kwargs for the Dense layer for all downsampling layers
        pooling : list of [int / tuple of int]
            the pooling window of all downsampling layers
        upsampling : list of [dict / list of dict]
            the kwargs for the Dense layer for all upsampling layers
        pooling_type : one of {"max", "avg"}
            The type of pooling to perform
        upsampling_method : one of {"nearest", "interpolate"}
            The type of pooling to perform
        stacked : bool
            default value for "stacked" in the 'dense_layers' kwargs
        activation : str
            default value for "activation" in the 'dense_layers' kwargs
        dropout : float or None
            default value for "dropout" in the 'dense_layers' kwargs
        """
        assert len(downsampling) == len(pooling) == len(upsampling)
        super().__init__()
        encoder, decoder = [], []
        channels = []
        for dense, pool in zip(downsampling, pooling):
            channels.append(in_channels)
            down = self.DownsamplingNd(in_channels, dense, pool,
                                       pooling_type=pooling_type,
                                       stacked=stacked,
                                       dropout=dropout,
                                       activation=activation,
                                       padded=True)
            in_channels = down.out_channels(in_channels)
            encoder.append(down)
        for dense, down, stack_channels in zip(upsampling, encoder[::-1],
                                               channels[::-1]):
            up = self.UpsamplingNd(in_channels+stack_channels,
                                   dense,
                                   down.downsampling_factor,
                                   upsampling_method=upsampling_method,
                                   stacked=stacked,
                                   dropout=dropout,
                                   activation=activation,
                                   padded=True)
            in_channels = up.out_channels(in_channels+stack_channels)
            decoder.append(up)
        self.encoder = self.EncoderNd.from_layers(encoder)
        self.decoder = self.DecoderNd.from_layers(decoder)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        stack = []
        for stage in self.encoder.stages:
            stack.append(X)
            X = stage(X)
        for stage, Xstack in zip(self.decoder.stages, stack[::-1]):
            X = stage(X, Xstack)
        return X

    def shape_out(self, shape_in: list) -> list:
        shape = shape_in
        shape = self.encoder.shape_out(shape)
        shape = self.decoder.shape_out(shape)
        return shape

    def shape_in(self, shape_out: list) -> list:
        shape = shape_out
        shape = self.decoder.shape_in(shape)
        shape = self.encoder.shape_in(shape)
        return shape

    def out_channels(self, in_channels):
        channels = []
        for stage in self.encoder.stages:
            channels.append(in_channels)
            in_channels = stage.out_channels(in_channels)
        for stage, stack in zip(self.decoder.stages, channels[::-1]):
            in_channels = stage.out_channels(in_channels+stack)
        return in_channels

    @property
    def UpsamplingNd(self):
        return self.DecoderNd.UpsamplingNd

    @property
    def DownsamplingNd(self):
        return self.EncoderNd.DownsamplingNd

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "encoder": self.encoder.dump,
                "decoder": self.decoder.dump}


class UNet1d(UNet):

    EncoderNd = Encoder1d
    DecoderNd = Decoder1d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UNet2d(UNet):

    EncoderNd = Encoder2d
    DecoderNd = Decoder2d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import json
import pathlib


class Model:

    @classmethod
    def load(cls, file: str):
        file = pathlib.Path(file)
        if not file.is_file():
            raise FileNotFoundError("The file '{file}' does not exist")
        with open(file) as json_file:
            dump = json.load(json_file)
        return cls.from_dump(dump)

    def save(self, file: str, overwrite: bool = False):
        file = pathlib.Path(file)
        path = file.parent
        if not path.is_dir():
            raise ValueError(f"The directory '{path}' does not exist")
        if not(overwrite) and file.exists():
            raise FileExistsError("The file '{file}' already exists,"
                                  " set 'overwrite=True' to overwrite.")
        with open(file, "w") as json_file:
            json.dump(self.dump, json_file)



import torch
import torch.nn.functional as F
from typing import Union


class Dropout(torch.nn.Module):
    """
    In training mode, randomly zero out each channel with probability 'p'
    (over the whole feature map if in 1d/2d)
    This is usefull to enforce redundancy in the channels of the neural
    network, thus increasing it's capacity to extrapolate to unseen data
    """

    @classmethod
    def from_dump(cls, dump: dict) -> "Dropout":
        assert dump["type"] == cls.__name__
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.p = dump["p"]

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "p": self.p}

    def __init__(self, p: Union[None, float]):
        super().__init__()
        self.p = p

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        if self.p is None:
            return X
        return F.dropout2d(X, self.p, training=self.training)



import torch
import pandas as pd
import numpy as np
from typing import List, Iterable, Union


def floats_to_tensor(arr: Iterable, device: torch.device) -> torch.Tensor:
    """converts an array of numerical values to a tensor of floats"""
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    return torch.tensor(arr, dtype=torch.float, device=device,
                        requires_grad=False)


def tensor_to_floats(tensor: torch.Tensor) -> np.ndarray:
    """converts a torch.Tensor to a numpy.ndarray of doubles"""
    assert tensor.dtype == torch.float
    return tensor.detach().cpu().numpy().astype(np.float64)


def longs_to_tensor(arr: Iterable, device: torch.device) -> torch.Tensor:
    """converts an array of numerical values to a tensor of longs"""
    return torch.tensor(arr, dtype=torch.long, device=device,
                        requires_grad=False)


def tensor_to_longs(tensor: torch.Tensor) -> list:
    """converts an array of numerical values to a tensor of longs"""
    assert tensor.dtype == torch.long
    return tensor.detach().cpu().numpy()


def images_to_tensor(images: Iterable[np.ndarray],
                     device: torch.device) -> torch.Tensor:
    """Converts a list of images to a tensor"""
    assert all(im.dtype == np.uint8 for im in images)
    # Numpy images are of shape (height, width, channel) or (height, width)
    # but pytorch expects (channels, height width)
    images = [np.moveaxis(im if len(im.shape) == 3 else im[:, :, None],
                          -1, 0).astype(float)
              for im in images]
    return floats_to_tensor(images, device)


def tensor_to_images(tensor: torch.Tensor,
                     colors: Union[np.ndarray, None] = None
                     ) -> np.ndarray:
    """
    Converts a tensor of long to a list of images
    If 'colors' is not None, tensor must contain indexes to the
    color for each pixel.
    Otherwise it must be a tensor of float valued images between 0 and 255.
    """
    if colors is None:
        assert tensor.dtype == torch.float
        arr = np.round(tensor_to_floats(tensor))  # round results
        arr = np.clip(arr, 0, 255).astype("uint8")  # converts to uint8
        if arr.shape[1] == 1:  # grayscale images
            return arr[:, 0, :, :]
        elif arr.shape[1] in [3, 4]:  # RGB or RGBA image
            return np.moveaxis(arr, 1, -1)
    else:
        assert tensor.dtype == torch.long
        return colors[tensor_to_longs(tensor)]


def classes_to_tensor(input: Iterable[str],
                         classes: List[str],
                         device: torch.device) -> torch.Tensor:
    """
    converts a list of categorical variables to tensor
    'classes' must be a list of unique possible classes.
    The tensor contains for each input the index of the category.
    """
    assert isinstance(classes, list)
    return longs_to_tensor([classes.index(i) for i in input], device=device)


def tensor_to_index(tensor: torch.tensor) -> np.ndarray:
    """Converts a tensor to an array of category index"""
    return tensor_to_longs(torch.argmax(tensor, dim=1))


def tensor_to_classes(tensor: torch.Tensor,
                         classes: List[str]) -> List[str]:
    """Converts a tensor of category indexes to str category"""
    indexes = tensor_to_index(tensor)
    return np.array(classes)[indexes]


def dataframe_to_tensor(df: pd.DataFrame,
                        x: List[str],
                        device: torch.device) -> torch.Tensor:
    """Converts a dataframe of numerical values to tensor"""
    assert all(np.issubdtype(df[var].dtype, np.number) for var in x)
    arr = df[x].to_numpy(dtype=np.float32)
    return floats_to_tensor(arr, device=device)


def tensor_to_probabilities(tensor: torch.Tensor,
                            classes: List[str]) -> pd.DataFrame:
    """
    Converts the raw output of a classifier neural network
    to a dataframe of class probability for each observation
    """
    arr = tensor_to_floats(torch.softmax(tensor, dim=-1))
    return pd.DataFrame(data=arr, columns=classes)


def segmented_to_tensor(images: np.ndarray, colors: Iterable,
                        device: torch.device) -> torch.Tensor:
    """
    Converts a segmented image to a tensor of long
    """
    if len(images.shape) == 4:  # Color image
        assert all(hasattr(c, "__iter__") for c in colors)
    elif len(images.shape) == 3:  # Grayscale image
        assert all(isinstance(c, int) for c in colors)
        images = np.expand_dims(images, -1)
        colors = [[c] for c in colors]
    else:
        raise RuntimeError("Unexpected shape of segmented images")
    masks = np.array([np.all(images == c, axis=3) for c in colors])
    if not masks.any(axis=0).all():
        raise RuntimeError("Found color associated to no class")
    return longs_to_tensor(np.argmax(masks, axis=0), device=device)



import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from typing import Any, Tuple, Iterable, List, Union


def split(data: Tuple[Any], frac: float = 0.2, shuffle: bool = True) -> tuple:
    """
    Splits the input data in two (train, test)

    Parameters
    ----------
    data : tuple
        Tuple of iterables
    frac : float
        The fraction of testing data
    shuffle : bool
        If True, the data is shuffled before splitting

    Returns
    -------
    tuple :
        the 'first' and 'second' tuples of data
    """
    L = len(data[0])
    indexes = np.random.permutation(L) if shuffle else np.arange(L)
    limit = int(round(frac * L))
    b = indexes[:limit]
    a = indexes[limit:]
    train = [_index(d, a) for d in data]
    test = [_index(d, b) for d in data]
    return tuple(train), tuple(test)


def kfold(data: Tuple[Any], k: int = 3, shuffle: bool = True) -> tuple:
    """
    Splits the input data into k-folds of (train, test) data

    Parameters
    ----------
    data : tuple
        Tuple of iterables
    k : int
        The number of folds to yield
    shuffle : bool
        If True, the data is shuffled before splitting


    Yields
    ------
    tuple :
        the (train, test) tuple of data
    """
    L = len(data[0])
    indexes = np.random.permutation(L) if shuffle else np.arange(L)
    indexes = np.split(indexes, k)
    for i in range(k):
        train = []
        for j, ind in enumerate(indexes):
            if j == i:
                test = ind
            else:
                train.extend(ind)
        yield _index(data, train), _index(data, test)


def MSE(predicted: np.ndarray, target: np.ndarray, weights=None):
    """Returns the Mean Squared Error of a regressor"""
    assert len(predicted) == len(target)
    SE = (predicted - target)**2
    if weights is not None:
        SE *= weights
    return np.mean(SE)


def RMSE(predicted: np.ndarray, target: np.ndarray, weights=None):
    """Returns the Root Mean Squared Error of a regressor"""
    return np.sqrt(MSE(predicted, target, weights=weights))


def R2(predicted: np.ndarray, target: np.ndarray, weights=None):
    """Returns the R² score of a regressor"""
    assert len(predicted) == len(target)
    SEres = (predicted - target)**2
    SEtot = (target - np.mean(target))**2
    if weights is not None:
        SEres *= weights
        SEtot *= weights
    return 1 - np.sum(SEres)/np.sum(SEtot)


def accuracy(predicted: np.ndarray, target: np.ndarray):
    """Returns the accuracy of a classifier"""
    assert len(predicted) == len(target)
    return sum([a == b for a, b in zip(predicted, target)])/len(predicted)


def plot_correlation(predicted: Iterable[float], target: Iterable[float],
                     ax: Union[None, matplotlib.axes.Axes] = None,
                     label: str = "_",
                     **kwargs):
    """
    Plots the correlation between prediction and target of a regressor

    Parameters
    ----------
    predicted : iterable of str
        the classes predicted by the model
    target : iterable of str
        the target to predict
    ax : None or matplotlib.axes.Axes
        The axes to draw on. If None a new window is created.
    label : str
        The legend of the data plotted. Ignored if starts with '_'.
    **kwargs : dict
        dict of keywords passed to 'plt.scatter'
    """
    assert len(predicted) == len(target)
    if ax is None:
        f, ax = plt.subplots(figsize=[5, 5])
    ax.scatter(target, predicted, label=label, **kwargs)
    points = np.concatenate([c.get_offsets() for c in ax.collections])
    inf, sup = points.min(), points.max()
    delta = sup - inf if sup != inf else 1
    sup += 0.05*delta
    inf -= 0.05*delta
    plt.plot([inf, sup], [inf, sup], color="k", zorder=0)
    ax.set_xlim([inf, sup])
    ax.set_ylim([inf, sup])
    ax.set_xlabel("target")
    ax.set_ylabel("predicted")
    ax.set_aspect("equal", "box")
    ax.legend()


def confusion_matrix(predicted: Iterable[str], target: Iterable[str],
                     classes: Union[None, List[str]] = None):
    """
    Returns the confusion matrix between prediction and target
    of a classifier

    Parameters
    ----------
    predicted : iterable of str
        the classes predicted by the model
    target : iterable of str
        the target to predict
    classes : None or list of str
        the unique classes to plot
        (can be a subset of the classes in 'predicted' and 'target')
        If None, the classes are infered from unique values from
        'predicted' and 'target'
    """
    assert len(predicted) == len(target)
    if classes is None:
        classes = np.unique(np.stack([predicted, target]))
    predicted = pd.Series(predicted).reset_index(drop=True)
    target = pd.Series(target).reset_index(drop=True)
    tab = pd.crosstab(predicted, target, normalize="all")
    for c in classes:
        if c not in tab.index:
            tab.loc[c] = 0
        if c not in tab.columns:
            tab[c] = 0
    return tab.loc[classes, classes]


def plot_confusion_matrix(*args, ax: Union[None, matplotlib.axes.Axes] = None,
                          cmap: str = "Greens", **kwargs):
    """
    Plots the confusion matrix between prediction and target
    of a classifier

    Parameters
    ----------
    *args : tuple
        args passed to 'confusion_matrix'
    ax : None or matplotlib.axes.Axes
        The axis to draw on. If None, a new window is created.
    cmap : str
        The name of the maplotlib colormap
    **kwargs : dict
        kwargs passed to 'confusion_matrix'
    """
    if ax is None:
        f, ax = plt.subplots()
    tab = confusion_matrix(*args, **kwargs)
    inf, sup = tab.min().min(), tab.max().max()
    ax.imshow(tab.to_numpy(), origin="lower", interpolation="nearest",
              cmap=cmap, vmin=inf, vmax=sup)
    ax.grid(False)
    ax.set_xticks(range(len(tab.columns)))
    ax.set_xticklabels(tab.columns, rotation=45)
    ax.set_xlabel("target")
    ax.set_yticks(range(len(tab.index)))
    ax.set_yticklabels(tab.index, rotation=45)
    ax.set_ylabel("predicted")
    for y, cy in enumerate(tab.index):
        for x, cx in enumerate(tab.columns):
            val = tab.loc[cy, cx]
            if val >= 0.01:
                color = "white" if (val - inf)/(sup - inf) > 0.5 else "black"
                ax.text(x, y, f"{val:.2f}", va='center', ha='center',
                        color=color)


def _index(data: Any, at: np.ndarray):
    """Indexes an input data. Method depends on it's type"""
    if data is None:
        return None
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        return data.iloc[at]
    elif isinstance(data, np.ndarray):
        return data[at]
    elif isinstance(data, list):
        return [data[i] for i in at]
    else:
        raise RuntimeError(f"'{type(data)}' can't be indexed")



import requests
import pathlib
from typing import List


def download(directory: str, file_name: str, url: str):
    """
    Download a file from the given url to the disk.
    If the directory does not exists raise an error.
    If the file already exists skip it.

    Parameters
    ----------
    directory : str
        directory in which the file is saved
    file_name : str
        name of the file
    url : str
        url to download it from
    """
    # test if path are valid
    directory = pathlib.Path(directory)
    if not directory.is_dir():
        raise NotADirectoryError(f"The path '{directory}' is not an existing directory")
    path = directory / file_name
    if path.is_file():
        print(f"skipping file '{file_name}' as it already exists", flush=True)
        return
    # make the request
    session = requests.Session()
    response = session.get(direct_url(url), stream=True)
    if response.status_code >= 400:
        raise RuntimeError(f"http error: {response.status_code}")
    # get a confirmation token for large files
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            response = session.get(direct_url(url), params={'confirm': value},
                                   stream=True)
            break
    # save on disk
    CHUNK_SIZE = 32768
    with open(path, "wb") as f:
        print(f"{file_name}: 0. kB", end="", flush=True)
        for i, chunk in enumerate(response.iter_content(CHUNK_SIZE)):
            f.write(chunk)
            n_bytes = (i+1)*CHUNK_SIZE / 8.
            for j, unit in enumerate(['bytes', 'kB', 'MB', 'GB', 'TB']):
                if n_bytes < 10**(3*(j+1)):
                    break
            progress = n_bytes / 10**(3*j)
            print(f"\r{file_name}: {progress:.2g} {unit}"+" "*10,
                  end="", flush=True)
    print()


def direct_url(url: str) -> str:
    """
    Converts a googledrive 'share' url to a direct download url

    Parameters
    ----------
    url : str
        the link of of a shared googledrive file

    Returns
    -------
    str :
        the direct download url
    """
    id = url.split("/")[-2]
    return f"https://docs.google.com/uc?export=download&id={id}"


def downloads(directory: str, folder_name: str,
              file_names: List[str], urls: List[str]):
    """
    Download a series of files in a new folder.

    Parameters
    ----------
    directory : str
        the base directory where the dataset must be stored
    folder_name : str
        the folder to create in the directory
    file_names : list of str
        the list of file names to create
    urls : list of str
        the list of urls to download them from
    """
    # test if path are valid
    directory = pathlib.Path(directory)
    if not directory.is_dir():
        raise NotADirectoryError(f"'{directory}' is not an existing directory")
    directory = directory / folder_name
    if not directory.is_dir():
        directory.mkdir(parents=True, exist_ok=True)
    # download each file
    for file_name, url in zip(file_names, urls):
        download(directory, file_name, url)


def boston_housing(directory: str):
    """downloads 'boston_housing.csv' in the given directory"""
    download(directory, "boston_housing.csv", "https://drive.google.com/file/d/1fTWYixdKF4tWyhD3V-qCDSZmReN_6LzP/view?usp=sharing")


def iris(directory: str):
    """downloads 'iris.csv' in the given directory"""
    download(directory, "iris.csv", "https://drive.google.com/file/d/1S1AHfTBtnW1SxsMskRmUcnoDehMbCj0R/view?usp=sharing")


def titanic(directory: str):
    """downloads 'titanic.csv' in the given directory"""
    download(directory, "titanic.csv", "https://drive.google.com/file/d/1LYjbHW3wyJSMzGMMCmaOFNA_RIKqxRoI/view?usp=sharing")


def fashion_mnist(directory: str):
    """downloads the 'fashion MNIST' dataset in the given directory"""
    file_names = ["classes.txt", "test_images.npy", "test_labels.npy",
                  "train_images.npy", "train_labels.npy"]
    urls = ["https://drive.google.com/file/d/1hrQfvcIbpe48JUIS870Ubah0wAFJvcEI/view?usp=sharing",
            "https://drive.google.com/file/d/1wLuRtkRIb1z7A92V86CZxP4L80K2L7ip/view?usp=sharing",
            "https://drive.google.com/file/d/11mb6eMSGbsgvuEpDzdITMcAPlsJP6Oxt/view?usp=sharing",
            "https://drive.google.com/file/d/1Z9Tir5UwuiY4paqos4wSBYnkPZJIorOw/view?usp=sharing",
            "https://drive.google.com/file/d/1MRIs2UhmfiT29NkPCv0qclc5KtOm5kuG/view?usp=sharing"]
    downloads(directory, "fashion-MNIST", file_names, urls)


def cityscapes(directory: str):
    """downloads the 'cityscapes' dataset in the given directory"""
    file_names = ["class_fractions.json", "classes.json", "test_images.npy",
                  "test_segmented.npy", "train_images.npy",
                  "train_segmented.npy"]
    urls = ["https://drive.google.com/file/d/1WVUVJgoaovv-rHeIjztcY7J98oB3Hgmq/view?usp=sharing",
            "https://drive.google.com/file/d/1T89BJ0U9AtPXjfON7Du9Sbd1BniM2Q_T/view?usp=sharing",
            "https://drive.google.com/file/d/10Yfsfc_V0SMXh4cAlXHCnwJHtaWUvrK-/view?usp=sharing",
            "https://drive.google.com/file/d/1Gu9-hUZUhNYHsRc_PqwT-AupCuCBE5UU/view?usp=sharing",
            "https://drive.google.com/file/d/13DzRX1yUlDW8oXNiWEa-Bu02_myttlem/view?usp=sharing",
            "https://drive.google.com/file/d/1BmhXfQraa37rwsnvcub0x-FjPhir6RsJ/view?usp=sharing"]
    downloads(directory, "cityscapes", file_names, urls)



import torch
from .pooling import Pooling, Pooling1d, Pooling2d
from .dense import Dense, Dense1d, Dense2d
from typing import Union, List, Tuple


class Downsampling(torch.nn.Module):
    """
    A downsampling layer is a 'DenseNd' layer
    followed by a 'PoolingNd' layer.
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.dense = Dense.from_dump(dump["dense"])
        obj.pooling = Pooling.from_dump(dump["pooling"])

    def __init__(self, in_channels: int,
                 dense_layer: Union[List[dict], dict],
                 pooling_window: Union[int, Tuple[int, int]],
                 pooling_type: str,
                 **kwargs):
        """
        Parameters
        ----------
        in_channels : int
            the number of channels of the input
        dense_layer : dict, or list of dict
            the parameters of all layers of the 'DenseNd'
        pooling_window : int, or tuple of int
            the size of the pooling window
        pooling_type : 'max' or 'avg'
            The type of pooling
        **kwargs
            additional kwargs passed to DenseNd
        """
        super().__init__()
        dense = self.DenseNd(in_channels, dense_layer, **kwargs)
        pooling = self.PoolingNd(pooling_window, pooling_type)
        self.dense = dense
        self.pooling = pooling

    def forward(self, X: torch.tensor) -> torch.Tensor:
        X = self.dense(X)
        X = self.pooling(X)
        return X

    def shape_in(self, shape_out: list) -> list:
        return self.dense.shape_in(self.pooling.shape_in(shape_out))

    def shape_out(self, shape_in: list) -> list:
        return self.pooling.shape_out(self.dense.shape_out(shape_in))

    def in_channels(self, out_channels: int) -> int:
        return self.dense.in_channels(out_channels)

    def out_channels(self, in_channels: int) -> int:
        return self.dense.out_channels(in_channels)

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "dense": self.dense.dump,
                "pooling": self.pooling.dump}


class Downsampling1d(Downsampling):

    DenseNd = Dense1d
    PoolingNd = Pooling1d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def downsampling_factor(self) -> int:
        f = 1
        for activated in self.dense.layers:
            f *= activated.weighting.stride
        f *= self.pooling.pooling_window
        return f


class Downsampling2d(Downsampling):

    DenseNd = Dense2d
    PoolingNd = Pooling2d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def downsampling_factor(self) -> int:
        fh, fw = 1, 1
        for activated in self.dense.layers:
            h, w = activated.weighting.stride
            fh *= h
            fw *= w
        h, w = self.pooling.pooling_window
        return [fh*h, fw*w]



import torch
import torch.nn.functional as F
from typing import Tuple, List, Union


# def load_pooling(dump):
#     """returns the Pooling object loaded from a dump"""
#     return globals()[dump["type"]].from_dump(dump)


# class _Pooling:
#     """A template for pooling layers"""

#     @classmethod
#     def from_dump(cls, dump: dict) -> '_Pooling':
#         """returns a '_Pooling' layer from a dump"""
#         assert dump["type"] == cls.__name__
#         return cls(dump["kernel size"], stride=dump["stride"])

#     @property
#     def dump(self) -> dict:
#         ks = self.kernel_size
#         st = self.stride
#         return {"type": type(self).__name__,
#                 "kernel size": list(ks) if hasattr(ks, "__iter__") else ks,
#                 "stride": list(st) if hasattr(st, "__iter__") else st}

#     def shape_out(self, shape_in: list) -> list:
#         c = shape_in[0]
#         dims = shape_in[1:]
#         dims = [int((d - k)/s + 1) for d, k, s in
#                 zip(dims, self.kernel_size, self.stride)]
#         return [c] + dims

#     def shape_in(self, shape_out: list) -> list:
#         c = shape_out[0]
#         dims = shape_out[1:]
#         dims = [(d-1)*s + k for d, k, s in
#                 zip(dims, self.kernel_size, self.stride)]
#         return [c] + dims


# class MaxPool1d(torch.nn.MaxPool1d, _Pooling):
#     """A wrapper around torch.nn.MaxPool1d"""

#     def __init__(self, kernel_size: int, stride: int = 1):
#         """
#         Parameters
#         ----------
#         kernel_size : int
#             The size of the pooling window
#         stride : int
#             The stride of the pooling window
#         """
#         torch.nn.MaxPool1d.__init__(self, kernel_size, stride)


# class MaxPool2d(torch.nn.MaxPool2d, _Pooling):
#     """A wrapper around torch.nn.MaxPool2d"""

#     def __init__(self, kernel_size: Tuple[int, int], stride: int = 1):
#         """
#         Parameters
#         ----------
#         kernel_size : tuple of int
#             The (height, width) size of the pooling window
#         stride : tuple of int
#             The (height, width) stride of the pooling window
#         """
#         torch.nn.MaxPool1d.__init__(self, kernel_size, stride)


# class AvgPool1d(torch.nn.AvgPool1d, _Pooling):
#     """A wrapper around torch.nn.AvgPool1d"""

#     def __init__(self, kernel_size: int, stride: int = 1):
#         """
#         Parameters
#         ----------
#         kernel_size : int
#             The size of the pooling window
#         stride : int
#             The stride of the pooling window
#         """
#         torch.nn.AvgPool1d.__init__(self, kernel_size, stride)


# class AvgPool2d(torch.nn.AvgPool2d, _Pooling):
#     """A wrapper around torch.nn.AvgPool2d"""

#     def __init__(self, kernel_size: Tuple[int, int], stride: int = 1):
#         """
#         Parameters
#         ----------
#         kernel_size : tuple of int
#             The (height, width) size of the pooling window
#         stride : tuple of int
#             The (height, width) stride of the pooling window
#         """
#         torch.nn.AvgPool2d.__init__(self, kernel_size, stride)


# class _OverallPool(torch.nn.Module):
#     """
#     Over all pooling transforms a tensor of shape (N, C, ...)
#     to a tensor of shape (N, C)
#     """

#     @classmethod
#     def from_dump(cls, dump: dict) -> 'OverallPool1d':
#         assert dump["type"] == cls.__name__
#         return cls(dump["pooling type"])

#     def __init__(self, pooling_type: str, dimension: str):
#         """
#         Parameters
#         ----------
#         pooling type : one of {"Max", "Avg"}
#             the type of pooling to perform
#         dimension : one of {"1d", "2d"}
#             the dimension of the tensor to treat
#         """
#         super().__init__()
#         self.pooling_type = pooling_type
#         self.func = getattr(F, pooling_type.lower()+"_pool"+dimension)

#     def forward(self, X: torch.Tensor) -> torch.Tensor:
#         return self.func(X, X.shape[2:]).view(X.shape[:2])

#     @property
#     def dump(self):
#         return {"type": type(self).__name__,
#                 "pooling type": self.pooling_type}

#     def shape_in(self, shape_out: int) -> List[int]:
#         return [shape_out, 1, 1]

#     def shape_out(self, shape_in: List[int]) -> int:
#         return shape_in[0]


# class OverallPool1d(_OverallPool):

#     def __init__(self, pooling_type: str):
#         """
#         Parameters
#         ----------
#         pooling type : one of {"Max", "Avg"}
#             the type of pooling to perform
#         """
#         super().__init__(pooling_type, "1d")


# class OverallPool2d(_OverallPool):

#     def __init__(self, pooling_type: str):
#         """
#         Parameters
#         ----------
#         pooling type : one of {"Max", "Avg"}
#             the type of pooling to perform
#         """
#         super().__init__(pooling_type, "2d")


class Pooling(torch.nn.Module):
    """
    A pooling layer reduces the spatial dimension of a feature map
    by performing max or average pooling over a window for each channel
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        cls = globals()[dump["type"]]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.pooling_window = tuple(dump["pooling window"])
        obj.pooling_type = dump["pooling type"]

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "pooling window": [s for s in self.pooling_window],
                "pooling type": self.pooling_type}

    def __init__(self, pooling_window, pooling_type):
        super().__init__()
        self.pooling_type = pooling_type
        self.pooling_window = pooling_window

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        if self.pooling_window is None:
            window = tuple(X.shape[2:])
            return self.function(X, window).view(X.shape[:2])
        else:
            return self.function(X, self.pooling_window)

    def shape_out(self, shape_in: list) -> Union[list, int]:
        if self.pooling_window is None:
            return []
        return [int((d - w)/w + 1) for d, w in
                zip(shape_in, self.pooling_window)]

    def shape_in(self, shape_out: Union[list, int]) -> list:
        if self.pooling_window is None:
            return []
        return [(d-1)*w + w for d, w in
                zip(shape_out, self.pooling_window)]

    @property
    def function(self):
        return self.functions[self.pooling_type]


class Pooling1d(Pooling):

    functions = {"max": F.max_pool1d, "avg": F.avg_pool1d}

    def __init__(self, pooling_window: Union[int, None],
                 pooling_type: str = "max"):
        """
        Parameters:
        -----------
        pooling_window : int or None
            The size of the pooling window
            Or None if overall pool
        pooling_type : one of {"max", "avg"}
            The type of pooling to perform
        """
        super().__init__(pooling_window, pooling_type)


class Pooling2d(Pooling):

    functions = {"max": F.max_pool2d, "avg": F.avg_pool2d}

    def __init__(self, pooling_window: Union[Tuple[int, int], None],
                 pooling_type: str = "max"):
        """
        Parameters:
        -----------
        pooling_window : tuple of int or None
            The (height, width) of the pooling window
            Or None if overall pool
        pooling_type : one of {"max", "avg"}
            The type of pooling to perform
        """
        super().__init__(pooling_window, pooling_type)



import torch
import torch.nn.functional as F
from typing import Union, Tuple


class Unpooling(torch.nn.Module):
    """
    An unpooling layer increases spatial dimensions of a feature map
    by upscaling it using linear/bilinear interpolation
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        cls = globals()[dump["type"]]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.factor = tuple(dump["factor"])
        return obj

    def __init__(self, factor: Union[int, Tuple[int]],
                 method: str):
        """
        Parameters:
        -----------
        factor : int, or Tuple of int
            the upsampling factor
        method : one of {'nearest', 'interpolate'}
            the method used to
        """
        assert method in ["nearest", "interpolate"]
        super().__init__()
        self.factor = factor
        self.method = method

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "factor": [f for f in self.factor],
                "method": self.method}


class Unpooling1d(Unpooling):

    def __init__(self, factor: int = 2,
                 method: str = "nearest"):
        super().__init__(factor, method)

    def forward(self, X):
        mode = "linear" if self.method == "interpolate" else "nearest"
        return F.interpolate(X, scale_factor=self.factor,
                             mode=mode)


class Unpooling2d(Unpooling):

    def __init__(self, factor: Tuple[int, int] = (2, 2),
                 method: str = "nearest"):
        super().__init__(factor, method)

    def forward(self, X):
        mode = "bilinear" if self.method == "interpolate" else "nearest"
        align = False if self.method == "interpolate" else None
        return F.interpolate(X, scale_factor=self.factor,
                             mode=mode,
                             align_corners=align)



import torch
from typing import Union


def load_batchnorm(dump):
    """returns the BatchNorm1d or BatchNorm2d object loaded from a dump"""
    return globals()[dump["type"]].from_dump(dump)


class BatchNorm:
    """A template for batch norm layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> 'BatchNorm':
        """returns a 'BatchNorm' layer from a dump"""
        cls = globals()[dump["type"]]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.momentum = dump["momentum"]
        obj.affine = dump["affine"]
        obj.eps = dump["eps"]
        obj.running_mean = torch.tensor(dump["running mean"],
                                        dtype=torch.float)
        obj.running_var = torch.tensor(dump["running var"],
                                       dtype=torch.float)
        if obj.affine:
            obj.weight = torch.nn.Parameter(torch.tensor(dump["weight"],
                                                         dtype=torch.float))
            obj.bias = torch.nn.Parameter(torch.tensor(dump["bias"],
                                                       dtype=torch.float))
        return obj

    @property
    def dump(self) -> dict:
        w = self.weight.tolist() if self.affine else None
        b = self.bias.tolist() if self.affine else None
        return {"type": type(self).__name__,
                "num features": self.num_features,
                "running mean": self.running_mean.tolist(),
                "running var": self.running_var.tolist(),
                "momentum": self.momentum,
                "eps": self.eps,
                "affine": self.affine,
                "weight": w,
                "bias": b}

    def undo(self, X: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverse transform to a tensor

        Parameters
        ----------
        X : torch.Tensor
            the normalized tensor

        Returns
        -------
        torch.Tensor :
            The un-normalized tensor
        """
        return X*(self.running_var+self.eps)**0.5 + self.running_mean


class BatchNorm1d(torch.nn.BatchNorm1d, BatchNorm):
    """A wrapper around torch.nn.BatchNorm1d"""

    def __init__(self, num_features: int,
                 momentum: Union[float, None] = 0.1,
                 affine: bool = False,
                 eps: float = 1.0E-5):
        """
        Parameters
        ----------
        num_features : int
            the number of channels to normalize
        momentum : float of None
            the update factor for the running mean and variance:
            x_run = x_run*(1-momentum) + x_new*momentum
            if None, the cumulative average and variance are calculated instead
        affine : bool
            whether to add learnable weight and bias after normalization
        eps : float
            A small constant used to avoid dividing by 0 when normalizing
        """
        torch.nn.BatchNorm1d.__init__(self, num_features,
                                      momentum=momentum,
                                      affine=affine,
                                      eps=eps)


class BatchNorm0d(BatchNorm1d):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class BatchNorm2d(torch.nn.BatchNorm2d, BatchNorm):
    """A wrapper around torch.nn.BatchNorm2d"""

    def __init__(self, num_features: int,
                 momentum: Union[float, None] = 0.1,
                 affine: bool = False,
                 eps: float = 1.0E-5):
        """
        Parameters
        ----------
        num_features : int
            the number of channels to normalize
        momentum : float of None
            the update factor for the running mean and variance:
            x_run = x_run*(1-momentum) + x_new*momentum
            if None, the cumulative average and variance are calculated instead
        affine : bool
            whether to add learnable weight and bias after normalization
        eps : float
            A small constant used to avoid dividing by 0 when normalizing
        """
        torch.nn.BatchNorm2d.__init__(self, num_features,
                                      momentum=momentum,
                                      affine=affine,
                                      eps=eps)



from .activated import Activated0d, Activated1d, Activated2d
from .batch_norm import BatchNorm0d, BatchNorm1d, BatchNorm2d
from .decoder import Decoder1d, Decoder2d
from .dense import Dense0d, Dense1d, Dense2d
from .downsampling import Downsampling1d, Downsampling2d
from .dropout import Dropout
from .encoder import Encoder1d, Encoder2d
from .padding import Padding1d, Padding2d
from .pooling import Pooling1d, Pooling2d
from .u_net import UNet1d, UNet2d
from .unpooling import Unpooling1d, Unpooling2d
from .upsampling import Upsampling1d, Upsampling2d
from .weighting import Linear, Conv1d, Conv2d



import torch
import numpy as np
from typing import Union, List, Tuple
from .layers import Linear, BatchNorm2d
from .layers import Encoder2d, Pooling2d, Dense0d
from .conversions import floats_to_tensor, tensor_to_index
from .conversions import classes_to_tensor, images_to_tensor
from .neural_network_classifier import NeuralNetworkClassifier
from .loss_functions import cross_entropy


class ImageClassifierModule(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        obj.classes = dump["classes"]
        obj.input_norm = BatchNorm2d.from_dump(dump["input norm"])
        obj.encoder = Encoder2d.from_dump(dump["encoder"])
        obj.final_pool = Pooling2d.from_dump(dump["final pool"])
        obj.dense = Dense0d.from_dump(dump["dense"])
        obj.output = Linear.from_dump(dump["output"])
        return obj

    def __init__(self, in_channels: int,
                 classes: List[str],
                 convolutions: Union[List[dict], List[List[dict]]],
                 pooling: List[Tuple[int, int]],
                 dense: List[dict],
                 pooling_type: str = "max",
                 padded: bool = True,
                 activation: str = "relu",
                 stacked: bool = False,
                 dropout: Union[float, None] = None):
        """
        Parameters
        ----------
        in_channels : int
            the number of channels in the input images
        class : list of str
            the unique classes the model can predict
        convolutions : list of [dict / list of dict]
            the kwargs for the 'Activated2d' layers for all 'downsampling'
        pooling : list of [int / tuple of int]
            the pooling window of all downsampling layers
            (excepted the last one which is an overall pool)
        dense : list of dict
            the kwargs for the 'Activated0d' of the final 'Dense1d' layer
        pooling_type : one of {'max', 'avg'}
            the type of pooling
        padded : bool
            the default value for the 'padded' key of the kwargs
        activation : str
            the default value for the 'activation' key of the kwargs
        stacked : bool
            the default value for the 'stacked' key of the kwargs
        dropout : float or None
            the default value for the 'dropout' key of the kwargs
        """
        super().__init__()
        assert len(pooling) == len(convolutions) - 1
        self.classes = list(classes)
        self.input_norm = BatchNorm2d(in_channels)
        self.encoder = Encoder2d(in_channels, convolutions, pooling+[None],
                                 pooling_type=pooling_type,
                                 padded=padded,
                                 activation=activation,
                                 stacked=stacked,
                                 dropout=dropout)
        in_channels = self.encoder.out_channels(in_channels)
        self.dense = Dense0d(in_channels, dense, activation=activation,
                             stacked=stacked, dropout=dropout)
        in_channels = self.dense.out_channels(in_channels)
        self.output = Linear(in_channels, len(classes))

    def forward(self, X: torch.Tensor):
        X = self.input_norm(X)
        X = self.encoder(X)
        X = self.dense(X)
        return self.output(X)

    def data_to_tensor(self, X: np.ndarray,
                       Y: Union[None, List[str]],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        x = images_to_tensor(X, self.device)
        y = None if Y is None else classes_to_tensor(Y, self.classes,
                                                     self.device)
        w = None if weights is None else floats_to_tensor(weights, self.device)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        indexes = tensor_to_index(tensor)
        return [self.classes[i] for i in indexes]

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[torch.Tensor, None]):
        return cross_entropy(y_pred, y_target, weights=weights,
                             class_weights=self.class_weights)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "classes": list(self.classes),
                "input norm": self.input_norm.dump,
                "encoder": self.encoder.dump,
                "dense": self.dense.dump,
                "output": self.output.dump}


class ImageClassifier(NeuralNetworkClassifier):

    ModuleType = ImageClassifierModule

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import torch
import numpy as np
from typing import Union, List, Tuple, Dict, Iterable
from .layers import BatchNorm2d, Conv2d
from .layers import UNet2d
from .conversions import floats_to_tensor, tensor_to_index
from .conversions import segmented_to_tensor, images_to_tensor
from .neural_network_classifier import NeuralNetworkClassifier
from .loss_functions import cross_entropy


class SemanticSegmenterModule(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        obj.colors = dump["colors"]
        obj.classes = dump["classes"]
        obj.input_norm = BatchNorm2d.from_dump(dump["input norm"])
        obj.UNet = UNet2d.from_dump(dump["U-net"])
        obj.output = Conv2d.from_dump(dump["output"])
        return obj

    def __init__(self, in_channels: int,
                 colors: Dict[str, Union[int, List[int]]],
                 downsampling: List[Union[dict, List[dict]]],
                 pooling: List[Tuple[int, int]],
                 upsampling: List[Union[dict, List[dict]]],
                 pooling_type: str = "max",
                 upsampling_method: str = "nearest",
                 activation: str = "relu",
                 stacked: bool = False,
                 dropout: Union[float, None] = None):
        """
        Parameters
        ----------
        in_channels : int
            The number of channels of the input
        colors : dict
            a dict of {class: color}
        downsampling : list of [dict / list of dict]
            the kwargs for the 'Activated2d' layers for all 'downsampling'
        pooling : list of [int / tuple of int]
            the pooling window of all downsampling layers
        upsampling : list of [dict / list of dict]
            the kwargs for the 'Activated2d' layers for all 'upsampling'
        pooling_type : one of {'max', 'avg'}
            the type of pooling
        upsampling_method : one of {'nearest', 'interpolate'}
            the method used for the unpooling layers
        activation : str
            the default value for the 'activation' key of the kwargs
        stacked : bool
            the default value for the 'stacked' key of the kwargs
        dropout : float or None
            the default value for the 'dropout' key of the kwargs
        """
        super().__init__()
        self.classes = [c for c in colors.keys()]
        self.colors = [colors[c] for c in self.classes]
        self.input_norm = BatchNorm2d(in_channels)
        self.u_net = UNet2d(in_channels, downsampling, pooling, upsampling,
                            pooling_type=pooling_type,
                            upsampling_method=upsampling_method,
                            activation=activation,
                            stacked=stacked,
                            dropout=dropout)
        in_channels = self.u_net.out_channels(in_channels)
        self.output = Conv2d(in_channels, len(self.classes), (1, 1))

    def forward(self, X: torch.Tensor):
        X = self.input_norm(X)
        X = self.u_net(X)
        return self.output(X)

    def data_to_tensor(self, X: Iterable[np.ndarray],
                       Y: Union[None, List[str]],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        x = images_to_tensor(X, self.device)
        y = None if Y is None else segmented_to_tensor(Y, self.colors,
                                                       self.device)
        w = None if weights is None else floats_to_tensor(weights, self.device)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        indexes = tensor_to_index(tensor)
        return np.array(self.colors)[indexes]

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[torch.Tensor, None]):
        return cross_entropy(y_pred, y_target, weights=weights,
                             class_weights=self.class_weights)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "classes": self.classes,
                "colors": self.colors,
                "input norm": self.input_norm.dump,
                "U-net": self.u_net.dump,
                "output": self.output.dump}


class SemanticSegmenter(NeuralNetworkClassifier):

    ModuleType = SemanticSegmenterModule

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
import setuptools
from pygmalion.info import __version__, __author__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pygmalion",
    version=__version__,
    author=__author__,
    author_email="benoitfamillefavier@gmail.com",
    description="A machine learning package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "pandas>=0.25",
        "numpy>=1.18",
        "torch>=1.2",
        "requests>=2.25",
        "matplotlib>=3.2"
    ],
    python_requires='>=3.6',
    url="https://github.com/BFavier/Pygmalion",
    packages=setuptools.find_packages(),
    classifiers=(                                 # Classifiers help people find your 
        "Programming Language :: Python :: 3",    # projects. See all possible classifiers 
        "License :: OSI Approved :: MIT License", # in https://pypi.org/classifiers/
        "Operating System :: OS Independent",
        "Environment :: GPU :: NVIDIA CUDA"
    ),
)



import torch
import numpy as np
import pandas as pd
from typing import Iterable
from .conversions import tensor_to_index, tensor_to_probabilities
from .neural_network import NeuralNetwork


class NeuralNetworkClassifier(NeuralNetwork):

    @classmethod
    def from_dump(cls, dump: dict) -> 'NeuralNetworkClassifier':
        obj = NeuralNetwork.from_dump.__func__(cls, dump)
        obj.class_weights = dump["class weights"]
        return obj

    def __init__(self, *args, class_weights=None, **kwargs):
        super().__init__(*args, **kwargs)
        if class_weights is None:
            self.class_weights = None
        else:
            self.class_weights = [class_weights[c] for c in self.classes]

    def index(self, X) -> np.ndarray:
        """
        Returns the category index for each observation

        Parameters
        ----------
        X : Any
            The input X of the model.
            it's type depend on the neural network type.
            see 'help(self.module)'

        Returns
        -------
        np.ndarray :
            An array of long, indexes of the 'self.category'
        """
        x, _, _ = self.module.data_to_tensor(X, None, None)
        return tensor_to_index(self.module(x))

    def probability(self, X) -> pd.DataFrame:
        """
        Return the category probabilities for each observations

        Parameters
        ----------
        X : Any
            The input X of the model.
            it's type depend on the neural network type.
            see 'help(self.module)'

        Returns
        -------
        pd.DataFrame :
            DataFrame of category probabilities
            where each column corresponds to a category
        """
        x, _, _ = self.module.data_to_tensor(X, None, None)
        return tensor_to_probabilities(self.module(x), self.classes)

    @property
    def class_weights(self):
        if self.module.class_weights is None:
            return None
        else:
            return self.module.class_weights.tolist()

    @class_weights.setter
    def class_weights(self, other: Iterable[float]):
        if other is not None:
            assert len(other) == len(self.classes)
            other = torch.tensor(other, dtype=torch.float,
                                 device=self.module.device)
        self.module.class_weights = other

    @property
    def classes(self):
        return self.module.classes

    @property
    def dump(self):
        d = super().dump
        d["class weights"] = self.class_weights
        return d



import torch
import torch.nn.functional as F
from typing import Union, Tuple
from .padding import Padding, Padding1d, Padding2d
from .weighting import Weighting, Linear, Conv1d, Conv2d
from .batch_norm import BatchNorm, BatchNorm1d, BatchNorm2d
from .dropout import Dropout


class Activated(torch.nn.Module):
    """
    An 'Activated' layer is a succession of:
        - Padding (optional, only for 1d/2d case)
        - Weighting (linear for 0d, or convolution for 1d/2d)
        - An activation function
        - Batch normalization
        - Dropout (during training, optional)
    """

    @classmethod
    def from_dump(cls, dump: dict):
        cls = globals()[dump["type"]]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        if "padding" in dump.keys():
            obj.padding = Padding.from_dump(dump["padding"])
        obj.weighting = Weighting.from_dump(dump["weighting"])
        obj.activation = dump["activation"]
        obj.normalization = BatchNorm.from_dump(dump["normalization"])
        obj.dropout = Dropout.from_dump(dump["dropout"])
        obj.stacked = dump["stacked"]
        return obj

    def shape_out(self, shape_in: list) -> list:
        shape_out = shape_in
        if hasattr(self, "padding"):
            shape_out = self.padding.shape_out(shape_in)
        shape_out = self.weighting.shape_out(shape_out)
        return shape_out

    def shape_in(self, shape_out: list) -> list:
        shape_in = self.weighting.shape_in(shape_out)
        shape_in = self.padding.shape_in(shape_in)
        return shape_in

    def out_channels(self, in_channels: int) -> int:
        assert in_channels == self.weighting.in_channels
        c = self.weighting.out_channels
        if self.stacked:
            c += self.weighting.in_channels
        return c

    def in_channels(self, out_channels: int) -> int:
        if self.stacked:
            assert out_channels == (self.weighting.out_channels
                                    + self.weighting.in_channels)
        else:
            assert out_channels == self.weighting.out_channels
        return self.weighting.in_channels

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        input = X
        if hasattr(self, "padding"):
            X = self.padding(X)
        X = self.weighting(X)
        X = self.function(X)
        X = self.normalization(X)
        X = self.dropout(X)
        if self.stacked:
            X = torch.cat([self.downsample(input), X], dim=1)
        return X

    @property
    def dump(self):
        d = {"type": type(self).__name__}
        if hasattr(self, "padding"):
            d["padding"] = self.padding.dump
        d["weighting"] = self.weighting.dump
        d["activation"] = self.activation
        d["normalization"] = self.normalization.dump
        d["dropout"] = self.dropout.dump
        d["stacked"] = self.stacked
        return d

    @property
    def function(self):
        if hasattr(torch, self.activation):
            return getattr(torch, self.activation)
        else:
            return getattr(F, self.activation)


class Activated0d(Activated):

    def __init__(self, in_channels: int,
                 channels: int = 8,
                 activation: str = "relu",
                 dropout: Union[None, float] = None,
                 bias: bool = True,
                 stacked: bool = False):
        super().__init__()
        self.weighting = Linear(in_channels, channels, bias=bias)
        self.normalization = BatchNorm1d(channels)
        self.activation = activation
        self.dropout = Dropout(dropout)
        self.stacked = stacked

    def out_channels(self, in_channels: int) -> int:
        assert in_channels == self.weighting.in_features
        c = self.weighting.out_features
        if self.stacked:
            c += self.weighting.in_features
        return c

    def in_channels(self, out_channels: int) -> int:
        if self.stacked:
            assert out_channels == (self.weighting.out_features
                                    + self.weighting.in_features)
        else:
            assert out_channels == self.weighting.out_features
        return self.weighting.in_channels

    def downsample(self, X: torch.Tensor) -> torch.Tensor:
        return X


class Activated1d(Activated):

    def __init__(self, in_channels: int,
                 channels: int = 8,
                 window: int = 3,
                 stride: int = 1,
                 activation: str = "relu",
                 dropout: Union[None, float] = None,
                 padded: bool = True,
                 bias: bool = True,
                 stacked: bool = False):
        super().__init__()
        if padded:
            left, right = window // 2, window - 1 - window // 2
            self.padding = Padding1d((left, right))
        self.weighting = Conv1d(in_channels, channels, kernel_size=window,
                                stride=stride, bias=bias)
        self.normalization = BatchNorm1d(channels)
        self.activation = activation
        self.dropout = Dropout(dropout)
        self.stacked = stacked

    def downsample(self, X: torch.Tensor) -> torch.Tensor:
        if not hasattr(self, "padding"):
            window = self.weighting.kernel_size
            left, right = window // 2, window - window // 2
            X = X[:, :, left:-right]
        if self.weighting.stride != 1:
            X = F.avg_pool1d(X, self.weighting.stride)
        return X


class Activated2d(Activated):

    def __init__(self, in_channels: int,
                 channels: int = 8,
                 window: Tuple[int, int] = (3, 3),
                 stride: Tuple[int, int] = (1, 1),
                 activation: str = "relu",
                 dropout: Union[None, float] = None,
                 padded: bool = True,
                 bias: bool = True,
                 stacked: bool = False):
        super().__init__()
        if padded:
            top, left = [w // 2 for w in window]
            bot, right = [w - 1 - w // 2 for w in window]
            self.padding = Padding2d((left, right, top, bot))
        self.weighting = Conv2d(in_channels, channels, kernel_size=window,
                                stride=stride, bias=bias)
        self.normalization = BatchNorm2d(channels)
        self.activation = activation
        self.dropout = Dropout(dropout)
        self.stacked = stacked

    def downsample(self, X: torch.Tensor) -> torch.Tensor:
        if not hasattr(self, "padding"):
            window = self.weighting.kernel_size
            top, left = [w // 2 for w in window]
            bot, right = [w - w // 2 for w in window]
            X = X[:, :, top:-bot, left:-right]
        if self.weighting.stride != (1, 1):
            X = F.avg_pool2d(X, self.weighting.stride)
        return X



import torch
from .unpooling import Unpooling, Unpooling1d, Unpooling2d
from .dense import Dense, Dense1d, Dense2d
from typing import Union, List, Tuple


class Upsampling(torch.nn.Module):
    """
    An upsampling layer is an 'UnpoolingNd' layer
    followed by a 'DenseNd' layer.
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.unpooling = Unpooling.from_dump(dump["unpooling"])
        obj.dense = Dense.from_dump(dump["dense"])

    def __init__(self, in_channels: int,
                 dense_layer: Union[List[dict], dict],
                 upsampling_factor: Union[int, Tuple[int, int]],
                 upsampling_method: str = "nearest",
                 **kwargs):
        """
        Parameters
        ----------
        in_channels : int
            the number of channels of the input
        dense_layer : dict, or list of dict
            the parameters of all layers of the 'DenseNd'
        upsampling_factor : int, or tuple of int
            the upsampling factor
        upsampling_method : one of {'nearest', 'interpolate'}
            the method used to unpool
        **kwargs
            additional kwargs passed to DenseNd
        """
        super().__init__()
        unpooling = self.UnpoolingNd(factor=upsampling_factor,
                                     method=upsampling_method)
        dense = self.DenseNd(in_channels, dense_layer, **kwargs)
        self.unpooling = unpooling
        self.dense = dense

    def forward(self, X: torch.tensor,
                Xstack: Union[torch.Tensor, None] = None) -> torch.Tensor:
        """
        Upsample X, optionnaly concatenate Xstack to it,
        then apply a dense layer.

        Parameters:
        -----------
        X : torch.Tensor
            the input of the model
        Xstack : torch.Tensor or None
            if a tensor is provided, the channels of Xstack are concatenated
            to the channels of X after the upsampling layer.
            This is usefull for UNet architectures

        Returns:
        -------
        torch.Tensor :
            result of the layer
        """
        X = self.unpooling(X)
        if Xstack is not None:
            X = self.concat(Xstack, X)
        X = self.dense(X)
        return X

    def shape_in(self, shape_out: list) -> list:
        return self.dense.shape_in(self.pooling.shape_in(shape_out))

    def shape_out(self, shape_in: list) -> list:
        return self.dense.shape_out(self.pooling.shape_out(shape_in))

    def in_channels(self, out_channels: int) -> int:
        return self.dense.in_channels(out_channels)

    def out_channels(self, in_channels: int) -> int:
        return self.dense.out_channels(in_channels)

    def concat(self, X1: torch.Tensor, X2: torch.Tensor) -> torch.Tensor:
        """
        return [X1, X2] concatenated along the channel axis
        if X2 is smaller than X1, it is padded with 0
        """
        padding = [[0, l1 - l2] for l1, l2 in
                   zip(X1.shape[2:], X2.shape[2:])]
        padding = sum(padding, [])
        if any(p > 0 for p in padding):
            X2 = torch.nn.functional.pad(X2, padding, value=0.)
        return torch.cat([X1, X2], dim=1)

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "unpooling": self.unpooling.dump,
                "dense": self.dense.dump}


class Upsampling1d(Upsampling):

    UnpoolingNd = Unpooling1d
    DenseNd = Dense1d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def factor(self) -> int:
        f = 1
        for activated in self.dense:
            f *= activated.weighting.stride
        f *= self.pooling.pooling_window
        return f


class Upsampling2d(Upsampling):

    UnpoolingNd = Unpooling2d
    DenseNd = Dense2d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def factor(self) -> Tuple[int, int]:
        fh, fw = 1, 1
        for activated in self.dense:
            h, w = activated.weighting.stride
            fh *= h
            fw *= w
        h, w = self.pooling.pooling_window
        return [fh*h, fw*w]



import torch
import torch.nn.functional as F
from typing import Union


def RMSE(y_pred: torch.Tensor, y_target: torch.Tensor,
         target_norm: torch.nn.Module,
         weights: Union[None, torch.Tensor] = None) -> torch.Tensor:
    """
    Returns the Root Mean Squared Error of the model.
    Each observation be optionnaly weighted

    Parameters
    ----------
    y_pred : torch.Tensor
        A Tensor of float of shape [N_observations]
        The values predicted by the model for eahc observations
    y_target : torch.Tensor
        A Tensor of float of shape [N_observations]
        The target values to be predicted by the model
    target_norm : torch.nn.Module
        A BatchNormNd module applied to the y_target
        before calculating the loss
    weights : None or torch.Tensor
        If None all observations are equally weighted
        Otherwise the squared error of each observation
        is multiplied by the given factor

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    y_target = target_norm(y_target)
    if weights is None:
        return torch.sqrt(F.mse_loss(y_pred, y_target))
    else:
        return torch.sqrt(torch.mean(weights * (y_pred - y_target)**2))


def cross_entropy(y_pred: torch.Tensor, y_target: torch.Tensor,
                  weights: Union[None, torch.Tensor] = None,
                  class_weights: Union[None, torch.Tensor] = None
                  ) -> torch.Tensor:
    """
    Returns the cross entropy error of the model.
    Each observation and each class be optionnaly weighted

    Parameters
    ----------
    y_pred : torch.Tensor
        A Tensor of float of shape [N_observations, N_classes, ...]
        The probability of each class for eahc observation
    y_target : torch.Tensor
        A Tensor of long of shape [N_observations, 1, ...]
        The index of the class to be predicted
    weights : None or torch.Tensor
        If None all observations are equally weighted
        Otherwise the squared error of each observation
        is multiplied by the given factor

    Returns
    -------
    torch.Tensor :
        the scalar value of the loss
    """
    if weights is None:
        return F.cross_entropy(y_pred, y_target, weight=class_weights)
    else:
        return F.nll_loss(F.log_softmax(y_pred, dim=1) * weights, y_target,
                          weight=class_weights)


def soft_dice_loss(y_pred: torch.Tensor, y_target: torch.Tensor,
                   weights: Union[None, torch.Tensor] = None,
                   class_weights: Union[None, torch.Tensor] = None):
    """
    A soft Dice loss for segmentation
    """
    # Dice coeff is numerator/denominator
    numerator = 2 * y_pred*y_target
    denominator = y_pred+y_target
    if weights is not None:
        numerator = numerator * weights
        denominator = denominator * weights
    # sum over feature map so that tnesor has shape [N_observations, N_classes]
    N_observations = y_pred.shape[0]
    N_classes = y_pred.shape[1]
    numerator = numerator.view(N_observations, N_classes, -1).sum(dim=-1)
    denominator = denominator.view(N_observations, N_classes, -1).sum(dim=-1)
    # Appy class weighting
    if class_weights is not None:
        numerator = numerator * class_weights
        denominator = denominator * class_weights
    # loss is  (1 - dice coeff) averaged over classes and observations
    dice_coeff = numerator / denominator
    return 1 - torch.mean(dice_coeff, dim=[0, 1])



import torch
from typing import Union, List, Tuple
from .upsampling import Upsampling, Upsampling1d, Upsampling2d


class Decoder(torch.nn.Module):
    """
    A Decoder is a succession of 'UpsamplingNd' layers.
    It increase the spatial dimensions of a feature map
    """

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.stages = torch.nn.ModuleList()
        for d in dump["stages"]:
            obj.stages.append(Upsampling.from_dump(d))
        return obj

    @classmethod
    def from_layers(cls, layers: List[object]):
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.stages = torch.nn.ModuleList()
        for layer in layers:
            assert isinstance(layer, cls.UpsamplingNd)
            obj.stages.append(layer)
        return obj

    def __init__(self, in_channels: int,
                 dense_layers: List[Union[dict, List[dict]]],
                 upsampling_factors: List[Union[int, Tuple[int, int]]],
                 upsampling_method: str = "nearest",
                 padded: bool = True,
                 stacked: bool = False,
                 activation: str = "relu",
                 dropout: Union[float, None] = None):
        """
        in_channels : int
            The number of channels of the input
        dense_layers : list of [dict / list of dict]
            the kwargs for the Dense layer for each
        upsampling_factors : list of [int / tuple of int]
            the window size of the pooling layers
            can be omited if pooling_type is None
        upsampling_method : one of {"nearest", "interpolate"}
            The type of pooling to perform
        padded : bool
            default value for "padded" in the 'dense_layers' kwargs
        stacked : bool
            default value for "stacked" in the 'dense_layers' kwargs
        activation : str
            default value for "activation" in the 'dense_layers' kwargs
        dropout : float or None
            default value for "dropout" in the 'dense_layers' kwargs
        """
        assert len(dense_layers) == len(upsampling_factors)
        super().__init__()
        self.stages = torch.nn.ModuleList()
        for dense_layer, factor in zip(dense_layers, upsampling_factors):
            stage = self.UpsamplingNd(in_channels, dense_layer,
                                      upsampling_factor=factor,
                                      upsampling_method=upsampling_method,
                                      padded=padded,
                                      stacked=stacked,
                                      activation=activation,
                                      dropout=dropout)
            self.stages.append(stage)
            in_channels = stage.out_channels(in_channels)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        for stage in self.stages:
            X = stage(X)
        return X

    def shape_out(self, shape_in: list) -> list:
        for stage in self.stages:
            shape_in = stage.shape_out(shape_in)
        return shape_in

    def shape_in(self, shape_out: list) -> list:
        for stage in self.stages[::-1]:
            shape_out = stage.shape_in(shape_out)
        return shape_out

    def in_channels(self, out_channels: int) -> int:
        for stage in self.stage[::-1]:
            out_channels = stage.in_channels(out_channels)
        return out_channels

    def out_channels(self, in_channels: int) -> int:
        for stage in self.stage:
            in_channels = stage.out_channels(in_channels)
        return in_channels

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "stages": [s.dump for s in self.stages]}


class Decoder1d(Decoder):

    UpsamplingNd = Upsampling1d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Decoder2d(Decoder):

    UpsamplingNd = Upsampling2d

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import torch
import pandas as pd
import numpy as np
from typing import List, Union, Iterable
from .layers import Dense0d, BatchNorm0d, Linear
from .conversions import dataframe_to_tensor, \
                         floats_to_tensor, tensor_to_floats
from .neural_network import NeuralNetwork
from .loss_functions import RMSE


class DenseRegressorModule(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        assert cls.__name__ == dump["type"]
        obj = cls.__new__(cls)
        obj.inputs = dump["inputs"]
        obj.input_norm = BatchNorm0d.from_dump(dump["input norm"])
        obj.dense = Dense0d.from_dump(dump["dense"])
        obj.output = Linear.from_dump(dump["output"])
        obj.target_norm = BatchNorm0d.from_dump(dump["target norm"])
        return obj

    def __init__(self, inputs: List[str],
                 hidden_layers: List[dict],
                 activation: str = "relu", stacked: bool = False,
                 dropout: Union[float, None] = None):
        """
        Parameters
        ----------
        inputs : list of str
            The name of the dataframe columns used as inputs
        hidden_layers : list of dict
            The kwargs of all 'Activated0d' layers
        activation : str
            default 'activation' parameter for 'Activated0d' layers
        stacked : bool
            default 'stacked' parameter for 'Activated0d' layers
        dropout : None or float
            default 'dropout' parameter for 'Activated0d' layers
        """
        super().__init__()
        self.inputs = list(inputs)
        in_channels = len(inputs)
        self.input_norm = BatchNorm0d(in_channels)
        self.dense = Dense0d(in_channels,
                             layers=hidden_layers,
                             activation=activation,
                             stacked=stacked,
                             dropout=dropout)
        in_channels = self.dense.out_channels(in_channels)
        self.output = Linear(in_channels, 1)
        self.target_norm = BatchNorm0d(1)

    def forward(self, x):
        x = self.input_norm(x)
        x = self.dense(x)
        return self.output(x)

    def data_to_tensor(self, X: Union[pd.DataFrame, Iterable],
                       Y: Union[None, np.ndarray],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        if isinstance(X, pd.DataFrame):
            x = dataframe_to_tensor(X, self.inputs, self.device)
        else:
            x = floats_to_tensor(X, self.device)
        y = None if Y is None else floats_to_tensor(Y, self.device).view(-1, 1)
        if weights is None:
            w = None
        else:
            w = floats_to_tensor(weights, self.device).view(-1, 1)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        return tensor_to_floats(self.target_norm.undo(tensor).view(-1))

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[None, torch.Tensor]) -> torch.Tensor:
        return RMSE(y_pred, y_target, self.target_norm, weights=weights)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "inputs": self.inputs,
                "input norm": self.input_norm.dump,
                "dense": self.dense.dump,
                "output": self.output.dump,
                "target norm": self.target_norm.dump}


class DenseRegressor(NeuralNetwork):

    ModuleType = DenseRegressorModule

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import torch
from typing import Tuple


def load_padding(dump):
    """returns the ConstantPad1d or ConstantPad2d object loaded from a dump"""
    return globals()[dump["type"]].from_dump(dump)


class _ConstantPad:
    """A template for constant pad layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> '_ConstantPad':
        """returns a '_ConstantPad' layer from a dump"""
        assert dump["type"] == cls.__name__
        return cls(dump["padding"], value=dump["value"])

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "padding": list(self.padding),
                "value": self.value}

    def shape_out(self, shape_in: list) -> list:
        c = shape_in[0]
        dims = shape_in[1:]
        before = self.padding[0::2]
        after = self.padding[1::2]
        dims = [b+d+a for d, b, a in
                zip(dims, before, after)]
        return [c] + dims

    def shape_in(self, shape_out: list) -> list:
        c = shape_out[0]
        dims = shape_out[1:]
        before = self.padding[0::2]
        after = self.padding[1::2]
        dims = [-b+d-a for d, b, a in
                zip(dims, before, after)]
        return [c] + dims


class ConstantPad1d(torch.nn.ConstantPad1d, _ConstantPad):
    """A wrapper around torch.nn.ConstantPad1d"""

    def __init__(self, padding: Tuple[int, int], value: float = 0.):
        """
        Parameters
        ----------
        padding : tuple of int
            The (left, right) size of padding
        value : float
            The value of the padding
        """
        torch.nn.ConstantPad1d.__init__(self, padding, value=value)


class ConstantPad2d(torch.nn.ConstantPad2d, _ConstantPad):
    """A wrapper around torch.nn.ConstantPad2d"""

    def __init__(self, padding: Tuple[int, int, int, int], value: float = 0.):
        """
        Parameters
        ----------
        padding : tuple of int
            The (left, right, top, bottom) size of padding
        value : float
            The value of the padding
        """
        torch.nn.ConstantPad2d.__init__(self, padding, value=value)


class Padding:
    """A template for constant pad layers"""

    @classmethod
    def from_dump(cls, dump: dict) -> object:
        if dump is None:
            return None
        cls = globals()[dump["type"]]
        obj = cls.__new__(cls)
        torch.nn.Module.__init__(obj)
        obj.padding = tuple([size for _, size in dump["margins"].items()])
        obj.value = dump["value"]
        return obj

    @property
    def dump(self) -> dict:
        L = len(self.padding)
        return {"type": type(self).__name__,
                "margins": {side: size for side, size in
                            zip(["left", "right", "top", "bottom"][:L],
                                self.padding)
                            },
                "value": self.value}

    def shape_out(self, shape_in: list) -> list:
        c = shape_in[0]
        dims = shape_in[1:]
        before = self.padding[0::2]
        after = self.padding[1::2]
        dims = [b+d+a for d, b, a in
                zip(dims, before, after)]
        return [c] + dims

    def shape_in(self, shape_out: list) -> list:
        c = shape_out[0]
        dims = shape_out[1:]
        before = self.padding[0::2]
        after = self.padding[1::2]
        dims = [-b+d-a for d, b, a in
                zip(dims, before, after)]
        return [c] + dims


class Padding1d(torch.nn.ConstantPad1d, Padding):

    def __init__(self, padding: Tuple[int, int], value: float = 0.):
        """
        Parameters
        ----------
        padding : tuple of int
            The (left, right) size of padding
        value : float
            The value of the padding
        """
        torch.nn.ConstantPad1d.__init__(self, padding, value=value)


class Padding2d(torch.nn.ConstantPad2d, Padding):

    def __init__(self, padding: Tuple[int, int, int, int], value: float = 0.):
        """
        Parameters
        ----------
        padding : tuple of int
            The (left, right, top, bottom) size of padding
        value : float
            The value of the padding
        """
        torch.nn.ConstantPad2d.__init__(self, padding, value=value)


if __name__ == "__main__":
    import IPython
    IPython.embed()
